/*
  bgmatrix.cpp
  Inspired by :
  Matrix.cpp - Max7219 LED Matrix library for Arduino & Wiring
  Copyright (c) 2006 Nicholas Zambetti.  All right reserved.

  You can use it at commented in Nicholas Zambetti files.
  https://github.com/lstoll/arduino-libraries/tree/master/Matrix
  https://github.com/lstoll/arduino-libraries/blob/master/Matrix/Matrix.h  

  Many improvements could be done, but I didn't want to spend to much time 
  developping this module.
==============================================================================*/

/******************************************************************************
 * Includes
 ******************************************************************************/

#include "bgmatrix.h"
#include <Arduino.h>

/******************************************************************************
 * Definitions
 ******************************************************************************/

// Matrix registers
#define REG_NOOP   0x00
#define REG_DIGIT0 0x01
#define REG_DIGIT1 0x02
#define REG_DIGIT2 0x03
#define REG_DIGIT3 0x04
#define REG_DIGIT4 0x05
#define REG_DIGIT5 0x06
#define REG_DIGIT6 0x07
#define REG_DIGIT7 0x08
#define REG_DECODEMODE  0x09
#define REG_INTENSITY   0x0A
#define REG_SCANLIMIT   0x0B
#define REG_SHUTDOWN    0x0C
#define REG_DISPLAYTEST 0x0F

/******************************************************************************
 * Constructors
 ******************************************************************************/

Matrix::Matrix( uint8_t pinData, uint8_t pinClock, uint8_t pinLoad, 
                byte* pabBuffer, word wBufSize_8 /* = 1 */, uint8_t bNbScreens /* = 1 */) {
//=========================================================================================
// Constructor
// Record the pins used
_pinData = pinData;
_pinClock = pinClock;
_pinLoad = pinLoad;
_screens = bNbScreens;
_wBufSize_8 = wBufSize_8;    // Size of the buffer, over 8 = number of symbols in the buffer
_wBufSize = wBufSize_8 * 8;  // Size of the buffer

// Set the used pins as Outputs
pinMode(_pinClock, OUTPUT);
pinMode(_pinData, OUTPUT);
pinMode(_pinLoad, OUTPUT);
digitalWrite(_pinLoad, LOW);   // Prepare for the next data transmission

// Memorise the screen buffer
_pabBuffer = pabBuffer;

// initialize registers
Clear();             // clear display
setScanLimit(0x07);  // use all rows/digits
setBrightness(0x0F); // maximum brightness
setRegister(REG_SHUTDOWN, 0x01);    // normal operation
setRegister(REG_DECODEMODE, 0x00);  // pixels not integers
setRegister(REG_DISPLAYTEST, 0x00); // not in test mode
} // Matrix

/******************************************************************************
 * MAX7219 SPI
 ******************************************************************************/

void Matrix::WriteRegister(byte bReg, byte bData, byte bMode /* = 0 */) {
//=======================================================================
// Write 16 bits in the MAX7219 register.
// The four High bits are irrelevant
// The four next bits tell where data should be written, or change an internal register
// bData contains the 8 bits of a row of a display
// bMode indicate where the LEFT is.
byte bX = 0; // to scan the bits of the register and of the data

// register
for (bX=7; bX<255; bX--) { // scan the 8 bits of the register
  digitalWrite(_pinClock, LOW);   // tick
  digitalWrite(_pinData, bitRead(bReg, bX));
  digitalWrite(_pinClock, HIGH);  // tock
  }

// Data
for (bX=0; bX<8; bX++) { // scan the 8 bits of the data
  digitalWrite(_pinClock, LOW);   // tick
  if ((bMode >> 1) % 2 == 0) digitalWrite(_pinData, bitRead(bData, bX));
  else                       digitalWrite(_pinData, bitRead(bData, 7-bX));
  digitalWrite(_pinClock, HIGH);  // tock
  }
} // WriteRegister

void Matrix::setRegister(uint8_t reg, uint8_t data) {
//===================================================
// sets register to a byte value for all screens
for (uint8_t i = 0; i < _screens; ++i) {
  WriteRegister(reg, data, 2);
  }
digitalWrite(_pinLoad, HIGH);  // latch in data
digitalWrite(_pinLoad, LOW); // end
} // setRegister

/******************************************************************************
 * MAX7219 Configuration
 ******************************************************************************/

// sets how many digits are displayed
void Matrix::setScanLimit(uint8_t value) {
//========================================
  setRegister(REG_SCANLIMIT, value & 0x07);
}

// sets brightness of the display
void Matrix::setBrightness(uint8_t value) {
//=========================================
  setRegister(REG_INTENSITY, value & 0x0F);
}

/******************************************************************************
 * Helper Functions
 ******************************************************************************/

void Matrix::WriteDot(uint8_t x, uint8_t y, uint8_t value) {
//==========================================================
// Set a LED on (if value != 0)  or  off (if value == 0)

if (!_pabBuffer) return;

// uint8_t's can't be negative, so don't test for negative x and y.
if (x >= _wBufSize || y >= 8) return;

uint8_t offset = x; // record x
x %= 8;             // make x relative to a single matrix
offset -= x;        // calculate buffer offset

x = 7 - x;

// record value in buffer
if (value) _pabBuffer[y + offset] |= 0x01 << x;
else       _pabBuffer[y + offset] &= ~(0x01 << x);
} // WriteDot

/******************************************************************************
 * User API
 ******************************************************************************/

void Matrix::Clear(void) {
//========================
// clears screens and buffers
//if (!_pabBuffer) return;

for (word jj = 0; jj < _wBufSize; ++jj) _pabBuffer[jj] = 0;
} // Clear

void Matrix::Update(int nShift /*= 0*/, byte bMode /*= 0*/ ) {
//============================================================
// Data from the buffer is writter to the LED screen.
// nShift  shift all the LED to the right if nShift > 0, to the left if nShift < 0
// bMode indiquate where the LEFT should be, where the TOP should be.

if (_screens <= 0) return; // nothing to do
byte bCol = 0;  // to scan the 8 column of a display
byte bDisp = 0; // to scan the displays
word wBuf = 0;  // to shift the data bits

int nShiftMod = nShift;
 
// Move  nShiftMod  into the interval [0, _wBufSize[
while (nShiftMod < 0) nShiftMod += _wBufSize;
nShiftMod = nShiftMod % _wBufSize;

byte bShiftMod8 = nShiftMod % 8;  // to shift the data bits
int nShiftDiv8 = _wBufSize_8 - (nShiftMod / 8);  // to shift the data bits

//Serial.print(nShiftMod);  Serial.print(F("  ..  "));  Serial.print(nShiftMod % 8);  Serial.print(F("  ..  "));  Serial.println(nShiftDiv8);
  
for (bCol=0; bCol<8; bCol++) { // 8 = number of LEDs in a column. This scan the 8 lines.
  for (bDisp=0; bDisp<_screens; bDisp++) { // scan the displays
    // Right Shift
    if ((bMode >> 1) % 2 == 0) 
       wBuf = word( _pabBuffer[bCol + 8*((7-1-bDisp+nShiftDiv8)%_wBufSize_8)], 
                    _pabBuffer[bCol + 8*((8-1-bDisp+nShiftDiv8)%_wBufSize_8)])>>bShiftMod8;
    else
       wBuf = word( _pabBuffer[bCol + 8*((7+bDisp+nShiftDiv8)%_wBufSize_8)], 
                    _pabBuffer[bCol + 8*((8+bDisp+nShiftDiv8)%_wBufSize_8)])>>bShiftMod8;
    
    if (bMode % 2 == 0) WriteRegister(8-bCol, lowByte(wBuf), bMode); 
    else                WriteRegister(1+bCol, lowByte(wBuf), bMode); 
    }
    
  digitalWrite(_pinLoad, HIGH);  // Latch in data
  digitalWrite(_pinLoad, LOW);   // Prepare for the next data transmission
  }
} // Update

